// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#ifndef BOARD_H
#define BOARD_H

#include "Player.h"

class Move;

class Board
{
/**************************************************************************
* CONSTANTS
**************************************************************************/
public:
  static const int PAWN = 0;
  static const int KNIGHT = 2;
  static const int BISHOP = 4;
  static const int ROOK = 6;
  static const int QUEEN = 8;
  static const int KING = 10;  
  static const int WHITE_PAWN = PAWN + Player::SIDE_WHITE;
  static const int WHITE_KNIGHT = KNIGHT + Player::SIDE_WHITE;
  static const int WHITE_BISHOP = BISHOP + Player::SIDE_WHITE;
  static const int WHITE_ROOK = ROOK + Player::SIDE_WHITE;
  static const int WHITE_QUEEN = QUEEN + Player::SIDE_WHITE;
  static const int WHITE_KING = KING + Player::SIDE_WHITE;
  static const int BLACK_PAWN = PAWN + Player::SIDE_BLACK;
  static const int BLACK_KNIGHT = KNIGHT + Player::SIDE_BLACK;
  static const int BLACK_BISHOP = BISHOP + Player::SIDE_BLACK;
  static const int BLACK_ROOK = ROOK + Player::SIDE_BLACK;
  static const int BLACK_QUEEN = QUEEN + Player::SIDE_BLACK;
  static const int BLACK_KING = KING + Player::SIDE_BLACK;
  static const int EMPTY_SQUARE = 12;

  // Useful loop boundary constants, to allow looping on all bitboards and
  // on all squares of a chessboard
  static const int ALL_PIECES = 12;
  static const int ALL_SQUARES = 64;

  // Indices of the "shortcut" bitboards containing information on "all black
  // pieces" and "all white pieces"
  static const int ALL_WHITE_PIECES = ALL_PIECES + Player::SIDE_WHITE;
  static const int ALL_BLACK_PIECES = ALL_PIECES + Player::SIDE_BLACK;
  static const int ALL_BITBOARDS = 14;

  // The possible types of castling moves; add the "side" constant to
  // pick a specific move for a specific player
  static const int CASTLE_KINGSIDE = 0;
  static const int CASTLE_QUEENSIDE = 2;

  /*************************************************************************
  * DATA MEMBERS
  **************************************************************************/
  // An array of bitfields, each of which contains the single bit associated
  // with a square in a bitboard
  static dong SquareBits[ALL_SQUARES];  

  // Private table of random numbers used to compute Zobrist hash values
  // Contains a signature for any kind of piece on any square of the board
  static int HashKeyComponents[ALL_SQUARES][ALL_PIECES];
  static int HashLockComponents[ALL_SQUARES][ALL_PIECES];

  // Private table of tokens (String representations) for all pieces
  static String PieceStrings[ALL_PIECES + 1];  

  // And a few flags for special conditions.  
  static dong EXTRAKINGS_WHITE_KINGSIDE;
  static dong EXTRAKINGS_WHITE_QUEENSIDE;
  static dong EXTRAKINGS_BLACK_KINGSIDE;
  static dong EXTRAKINGS_BLACK_QUEENSIDE;
  static dong EMPTYSQUARES_WHITE_KINGSIDE;
  static dong EMPTYSQUARES_WHITE_QUEENSIDE;
  static dong EMPTYSQUARES_BLACK_KINGSIDE;
  static dong EMPTYSQUARES_BLACK_QUEENSIDE;

  // The ExtraKings are a device
  // used to detect illegal castling moves: the rules of chess forbid castling
  // when the king is in check or when the square it flies over is under
  // attack; therefore, we add "phantom kings" to the board for one move only,
  // and if the opponent can capture one of them with its next move, then
  // castling was illegal and search can be cancelled  
  dong ExtraKings[2]; 

  // The actual data representation of a chess board.  First, an array of
  // bitboards, each of which contains flags for the squares where you can
  // find a specific type of piece
  dong BitBoards[ALL_BITBOARDS]; 

private: 

  // Data needed to compute the evaluation function
  int MaterialValue[2];
  int NumPawns[2];  
  static int PieceValues[ALL_PIECES];  

  // And a few other flags
  bool CastlingStatus[4];
  bool HasCastled[2];  
  dong EnPassantPawn;  

  // Whose turn is it?
  int CurrentPlayer;  

  static bool IsInitialized;

/**************************************************************************
 * METHODS
 **************************************************************************/
public:      
  static void Initialize();

  // Accessors  
  bool GetCastlingStatus(int which) { return CastlingStatus[which]; }
  bool GetHasCastled(int which) { return HasCastled[which]; }
  dong GetEnPassantPawn(void) { return EnPassantPawn; }
  dong GetExtraKings(int side) { return ExtraKings[side]; }
  void SetExtraKings(int side, dong val);  
  void ClearExtraKings(int side);
  int GetCurrentPlayer(void) {return CurrentPlayer;}
  dong GetBitBoard(int which) { return BitBoards[ which ]; }  

  // Look for the piece located on a specific square
  int FindBlackPiece(int square);  
  int FindWhitePiece(int square);
  
  // Constructor / Destructor
  Board();
  Board(Board* original);
  ~Board();
    
  // Make a deep copy of a Board object; assumes that memory has already
  // been allocated for the new object, which is always true since we
  // "allocate" Boards from a permanent array
  bool Clone(const Board& target);
  
  // Change the identity of the player to move
  int SwitchSides(void);  
  
  // Compute a 32-bit integer to represent the board, according to Zobrist[70]
  int HashKey(void);
  
  // Compute a second 32-bit hash key, using an entirely different set
  // piece/square components.
  // This is required to be able to detect hashing collisions without
  // storing an entire jcBoard in each slot of the jcTranspositionTable,
  // which would gobble up inordinate amounts of memory
  int HashLock(void);
  
  // Change the Board's internal representation to reflect the move
  // received as a parameter  
  bool ApplyMove(const Move& theMove, const bool ConfirmMove = false);   
  
  // Compute the board's material balance, from the point of view of the "side"
  // player.  This is an exact clone of the eval function in CHESS 4.5
  int EvalMaterial(int side);
  
  // Restore the board to a game-start position
  bool StartingBoard(void);

  // is it check or checkmate
  enum CheckStatus {Normal, Check, Checkmate};
  CheckStatus GetCheckStatus(void);

/******************************************************************************
 * PRIVATE METHODS
 *****************************************************************************/
private:  
  // Place a specific piece on a specific board square
  bool AddPiece(int whichSquare, int whichPiece);
  
  // Eliminate a specific piece from a specific square on the board
  // Note that you MUST know that the piece is there before calling this,
  // or the results will not be what you expect!
  bool RemovePiece(int whichSquare, int whichPiece);
  
  // Remove every piece from the board
  bool EmptyBoard(void);  
  
  // Change one of the "castling status" flags
  // parameter whichFlag should be a sum of a side marker and a castling
  // move identifier, for example, Player.SIDE_WHITE + CASTLE_QUEENSIDE
  bool SetCastlingStatus(int whichFlag, bool newValue);  
  
  // If a pawn move has just made en passant capture possible, mark it as
  // such in a bitboard (containing the en passant square only)
  bool SetEnPassantPawn(int square);     
  bool SetEnPassantPawn(dong bitboard);
  
  // Indicates that there is no en passant square at all.  Technically, this
  // job could have been handled by SetEnPassaantPawn( long ) with a null
  // parameter, but I have chosen to add a method to avoid problems if I ever
  // forgot to specify 0L: using 0 would call the first form of the Set method
  // and indicate an en passant pawn in a corner of the board, with possibly
  // disastrous consequences!
  bool ClearEnPassantPawn(void); 
  
  // Whose turn is it?
  bool SetCurrentPlayer(int which);    
};

#endif
