// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#ifndef MOVE_H
#define MOVE_H

class Move
{
public:
  // The different types of moves recognized by the game
  static const int MOVE_NORMAL = 0;
  static const int MOVE_CAPTURE_ORDINARY = 1;
  static const int MOVE_CAPTURE_EN_PASSANT = 2;
  static const int MOVE_CASTLING_KINGSIDE = 4;
  static const int MOVE_CASTLING_QUEENSIDE = 8;
  static const int MOVE_RESIGN = 16;
  static const int MOVE_STALEMATE = 17;
  static const int MOVE_CHECKMATE = 18;
  static const int MOVE_PROMOTION_KNIGHT = 32;
  static const int MOVE_PROMOTION_BISHOP = 64;
  static const int MOVE_PROMOTION_ROOK = 128;
  static const int MOVE_PROMOTION_QUEEN = 256;

  // A pair of masks used to split the promotion and the non-promotion part of
  // a move type ID
  static const int PROMOTION_MASK = 480;
  static const int NO_PROMOTION_MASK = 31;

  // Alphabeta may return an actual move potency evaluation, or an upper or
  // lower bound only (in case a cutoff happens).  We need to store this
  // information in the transposition table to make sure that a given
  // value is actually useful in given circumstances.
  static const int EVALTYPE_ACCURATE = 0;
  static const int EVALTYPE_UPPERBOUND = 1;
  static const int EVALTYPE_LOWERBOUND = 2;

  // A sentinel value used to identify Move fields without valid data
  static const int NULL_MOVE = -1;

  // The moving piece; one of the constants defined by jcBoard
  int MovingPiece;

  // The piece being captured by this move, if any; another jcBoard constant
  int CapturedPiece;

  // The squares involved in the move
  int SourceSquare, DestinationSquare;

  // A type ID: is this a regular move, a capture, a capture AND promotion from
  // Pawn to Rook, etc.  Move generation determines this, by definition; storing
  // it here avoids having to "re-discover" the information in jcBoard.ApplyMove
  // at the cost of a few bytes
  int MoveType;

  // An evaluation of the move's potency, either as a result of an alphabeta
  // search of some kind or of a retrieval in the transposition table
  int MoveEvaluation;
  int MoveEvaluationType;
  int SearchDepth;

  /*************************************************************************
   * PUBLIC METHODS
   *************************************************************************/
  Move() {this->Reset();}
  Move(const Move& original) ;
  ~Move() {};

  void Copy(const Move& target);    
  bool Equals(const Move& target);
  bool Reset(void);  
};

#endif
