// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "Application.h"
#include "CollisionManager.h"

// Singleton instance
template<> Application* Singleton<Application>::ms_Singleton = 0;

// Creation/destruction
Application::Application() : mRoot(0), mCollisionManager(0) {};

Application::~Application() 
{   
  delete mCollisionManager;
  mRoot->destroySceneManager(mSceneManager);
  delete mRoot;   
}

// Running application
void Application::run()
{
	if (!initialize())
		return;

	preStartRendering();  

	mRoot->startRendering();  
}

SceneManager* Application::getSceneManager()
{
	return mSceneManager;
}

RenderWindow* Application::getRenderWindow()
{
	return mWindow;
}

CEGUI::OgreRenderer* Application::getGUIRenderer()
{
  return mGUIRenderer;
}

CEGUI::System* Application::getGUISystem()
{
  return mGUISystem;
}

// Application setup
bool Application::initialize()
{
  // create a new OGRE root at the heart of everything
	mRoot = new Root();    

  // go through the resources file and load it all up
	setupResources();
	
	// Configure engine
	if (!configure())
		return false;

	// Initialize root and create render window
	mWindow = mRoot->initialise(true);

  // meh - generic'll do for now
	chooseSceneManager();

	// Set default mipmap level (NB: some APIs ignore this)
	TextureManager::getSingleton().setDefaultNumMipmaps(5);

  // Initialise, parse scripts etc  
  ResourceGroupManager::getSingleton().initialiseAllResourceGroups();

  // CEGUI setup
  Ogre::RenderTarget *mRenderTarget = mWindow;
  mGUIRenderer = &CEGUI::OgreRenderer::bootstrapSystem(*mRenderTarget);
  mGUISystem = CEGUI::System::getSingletonPtr();
  mGUISystem->setMouseMoveScaling(1.25); // adjust the speed

  // Mouse    
  CEGUI::SchemeManager::getSingleton().create((CEGUI::utf8*)"WindowsLook.scheme");  
  CEGUI::MouseCursor::getSingleton().setImage("WindowsLook", "MouseArrow");

  // create a new collision manager
  mCollisionManager = new CollisionManager();

	return true;
}

void Application::setupResources()
{
	// Load resource paths from config file
  ConfigFile cf;
  cf.load("resources.cfg");

  // Go through all sections & settings in the file
  ConfigFile::SectionIterator seci = cf.getSectionIterator();

  String secName, typeName, archName;
  while (seci.hasMoreElements())
  {
      secName = seci.peekNextKey();
      ConfigFile::SettingsMultiMap *settings = seci.getNext();
      ConfigFile::SettingsMultiMap::iterator i;
      for (i = settings->begin(); i != settings->end(); ++i)
      {
          typeName = i->first;
          archName = i->second;
          ResourceGroupManager::getSingleton().addResourceLocation(
              archName, typeName, secName);
      }
  }  
}

bool Application::configure()
{
	// Try to load existing configuration
	if (mRoot->restoreConfig())
		return true;
	// Show configuration dialog
	return mRoot->showConfigDialog();
}

void Application::chooseSceneManager()
{
	mSceneManager = mRoot->createSceneManager(ST_GENERIC);  
}
