// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "GameOGRE.h"
#include "State.h"
#include "StateManager.h"
#include "CPUvCPUState.h"
#include "PvCPUState.h"
#include "PvPState.h"
#include "MainMenuState.h"
#include "ColourState.h"
#include "OptionState.h"
#include "HelpState.h"
#include "GameMenuState.h"
#include "PromotionMenuState.h"
#include "GameOverState.h"
#include "GameAI.h"
#include "Hand.h"

// Singleton instance
template<> GameOGRE* Singleton<GameOGRE>::ms_Singleton = 0;

GameOGRE::GameOGRE(Camera* camera, Camera* reflectcamera, Camera* reflectcamera2, SceneManager* scenemanager, RenderWindow* window)
     : mCamera(camera), mReflectCamera(reflectcamera), mReflectCameraGround(reflectcamera2), mSceneManager(scenemanager), mWindow(window), mPaused(false)
{
	// Initialize camera
	mCameraTarget = Ogre::Vector3(0.0, 2.0, 15.0); //
	mCameraWantedYaw = 0; 
	mCameraWantedPitch = -60; 
	mCameraWantedDistance = 250.0;
	mCameraCorrection = 15.0;

	mCamera->setOrientation(Quaternion(Radian(Degree(mCameraWantedYaw)), Ogre::Vector3::UNIT_Y)
		* Quaternion(Radian(Degree(mCameraWantedPitch)), Ogre::Vector3::UNIT_X));
	mCamera->setPosition(mCameraTarget - mCameraWantedDistance*mCamera->getDirection());  

  // Setup animation default
  Animation::setDefaultInterpolationMode(Animation::IM_LINEAR);
  Animation::setDefaultRotationInterpolationMode(Animation::RIM_LINEAR);

	// Create game states
  mStates.push_back(new CPUvCPUState);
	mStates.push_back(new PvCPUState);	
  mStates.push_back(new PvPState);	

	// Create menu states
	mStates.push_back(new MainMenuState);
  mStates.push_back(new ColourState);	
  mStates.push_back(new OptionState);	
	mStates.push_back(new HelpState);	
	mStates.push_back(new GameMenuState);
  mStates.push_back(new PromotionMenuState);
	mStates.push_back(new GameOverState);	

	// Set up state stack	
  StateManager::getSingleton().pushState(CPUvCPUState::getSingletonPtr());
	StateManager::getSingleton().pushState(MainMenuState::getSingletonPtr());            

  // start the demo mode off 
  GameAI::getSingletonPtr()->Start();  
}

GameOGRE::~GameOGRE()
{  
	// Delete game states
  for (std::vector<State*>::iterator i = mStates.begin(); i != mStates.end(); ++i)  
		delete *i;	
}

Ray GameOGRE::CreateRay(void)
{
  CEGUI::Point tMousePos = CEGUI::MouseCursor::getSingleton().getPosition(); 
  tMousePos.d_x /= GameOGRE::getSingletonPtr()->mWindow->getWidth(); 
  tMousePos.d_y /= GameOGRE::getSingletonPtr()->mWindow->getHeight(); 
  Ogre::Ray mouseRay = mCamera->getCameraToViewportRay(tMousePos.d_x, tMousePos.d_y);        
  return mouseRay;
}

void GameOGRE::Update(Real dt)
{  
  // animate hands
  if (!mPaused)    
    Hand::UpdateHands(dt);

	// Update camera position
	Real t = dt * mCameraCorrection;
	Quaternion wantedOrientation;
	Real currentDistance, distance;

	// Clamp to one, so camera doesn't move too far
	if (t > 1.0)
		t = 1.0;

	// Create quaternion describing wanted orientation
	wantedOrientation = Quaternion(Radian(Degree(mCameraWantedYaw)), Ogre::Vector3::UNIT_Y)
		* Quaternion(Radian(Degree(mCameraWantedPitch)), Ogre::Vector3::UNIT_X);

	currentDistance = (mCamera->getPosition() - mCameraTarget).length();
	distance = currentDistance + t*(mCameraWantedDistance - currentDistance);

	// Move camera towards wanted position
	mCamera->setOrientation(Quaternion::Slerp(t, mCamera->getOrientation(), wantedOrientation, true));
	mCamera->setPosition(mCameraTarget - distance*mCamera->getDirection());

  mReflectCamera->setOrientation(mCamera->getOrientation());
  mReflectCamera->setPosition(mCamera->getPosition());

  mReflectCameraGround->setOrientation(mCamera->getOrientation());   
  mReflectCameraGround->setPosition(mCamera->getPosition());
}
