// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "HandState.h"
#include "AnimationBlender.h"
#include "Hand.h"
#include "Piece.h"

// default constructor
HandState::HandState(Hand* hand, const ActionType action, const int SourceSquare, const int DestinationSquare, 
                     const int CastlingSource, const int CastlingDestination, const String PromoType) 
  : mHand(hand), mAnimationBlender(0), mMoveState(0), mTargetPosition(Ogre::Vector3::ZERO), 
    mStartingPosition(Ogre::Vector3(140,-7.5,180)), mMovementVector(Ogre::Vector3::ZERO), mActionType(action), mStateType(Static), 
    mSourceSquare(SourceSquare), mDestinationSquare(DestinationSquare), mStateDuration(0),
    mMPSceneNode(0), mCPSceneNode(0), mMovingPiece(0), mCapturedPiece(0), mOffset(Ogre::Vector3::ZERO),
    mCastleFirst(false), mCastlingSource(CastlingSource), mCastlingDestination(CastlingDestination), mPromoType(PromoType), mEnPassantFirst(false)
{
    // create a new animation blender and set the initial animation
    mAnimationBlender = new AnimationBlender(mHand->GetEntityPtr());        
    mAnimationBlender->init("Tap2");

    // get some other handy pointers
    mSceneManager = mHand->GetSceneManagerPtr();
    mHandNode = mHand->GetSceneNodePtr();
    mHandEntity = mHand->GetEntityPtr();        
    mMovingPiece = Piece::GetPiece(mSourceSquare);  
    
    if (mActionType == Capture)
      mCapturedPiece = Piece::GetPiece(mDestinationSquare);    
}

HandState::~HandState()
{
  // get rid of the movement track
  if (mMoveState)    
  {    
    mSceneManager->destroyAnimation("HandTrack" + StringConverter::toString(mHand->GetColour()));
    mMoveState = 0;
  }
  // and the animation blender
  delete mAnimationBlender;
} 

void HandState::AttachPiece(const String bonename, Piece* piece, SceneNode*& pieceNode)
{
  // record the previous scene node so it can be re-attached later
  pieceNode = piece->GetSceneNode();    

  // remove the piece from the original node  
  pieceNode->detachObject(piece->GetEntity()); 
  mHandEntity->detachObjectFromBone(piece->GetEntity());      

  // attach the piece to the index finger bone      
  Bone* bone = mHandEntity->getSkeleton()->getBone(bonename);     

  // work out the offset orientation
  Quaternion offsetRotation = (bone->_getDerivedOrientation().Inverse() * pieceNode->getOrientation());

  // work out the offset position
  Ogre::Matrix3 boneRot, nodeRot;    

  Ogre::Vector3 nodePos = pieceNode->getPosition();
  Ogre::Vector3 handPos = mHandNode->getPosition(); 
  Ogre::Vector3 bonePos = bone->_getDerivedPosition();

  bone->_getDerivedOrientation().ToRotationMatrix(boneRot);
  mHandNode->getOrientation().ToRotationMatrix(nodeRot);

  Ogre::Vector3 offsetPosition = (nodePos - (handPos + (nodeRot * bonePos))); 
  offsetPosition = offsetPosition * (nodeRot * boneRot); 
  
  // finally, attach the object to the bone  
  mHandEntity->attachObjectToBone(bonename, piece->GetEntity(), offsetRotation, offsetPosition);           
}

void HandState::DetachPiece(Piece* piece, SceneNode*& pieceNode)
{
  // remove the piece from the bone
  mHandEntity->detachObjectFromBone(piece->GetEntity());    

  // re-attach the piece to it's original scene node  
  if (pieceNode != 0)
    pieceNode->attachObject(piece->GetEntity());
}

void HandState::BaseCreateMovementTrack(void)
{
  // make sure the scene node's rotation is taken into account
  Ogre::Matrix3 nodeRot;      
  mHandNode->_getDerivedOrientation().ToRotationMatrix(nodeRot); 

  mTargetPosition = (mStartingPosition * nodeRot);  

  // store the movement vector (in world coordinates) 
  // of the target position from the entities current position
  mMovementVector = mTargetPosition - mHandNode->getPosition();  
      
  // create the path to be followed   
  Animation* anim = mSceneManager->createAnimation("HandTrack" + StringConverter::toString(mHand->GetColour()), mStateDuration);

  // Spline it for nice curves
  anim->setInterpolationMode(Animation::IM_SPLINE);

  // Create a track to animate the hand
  NodeAnimationTrack* track = anim->createNodeTrack(0, mHandNode);

  // Setup keyframes
  TransformKeyFrame* key = track->createNodeKeyFrame(0); // startposition
  key->setTranslate(mHandNode->getPosition());
  key->setRotation(mHandNode->getOrientation());  
  
  key = track->createNodeKeyFrame(mStateDuration * 0.5);
  key->setTranslate(Ogre::Vector3(mTargetPosition.x - (mMovementVector.x * 0.5), mTargetPosition.y + 65, mTargetPosition.z - (mMovementVector.z * 0.5)));
  key->setRotation(mHandNode->getOrientation());    

  key = track->createNodeKeyFrame(mStateDuration);
  key->setTranslate(mTargetPosition);  
  key->setRotation(mHandNode->getOrientation()); 
}

void HandState::BaseStartAnimation(void)
{
  // Create a new animation state to track this
  mMoveState = mSceneManager->createAnimationState("HandTrack" + StringConverter::toString(mHand->GetColour()));
  mMoveState->setTimePosition(0);
  mMoveState->setEnabled(true);
  mMoveState->setLoop(false);
}

// default update
HandState::StateType HandState::BaseUpdate(Real dt)
{  
  // update the movement track if it is enabled
  if (mMoveState != 0)
  {
    if (mMoveState->getEnabled())
      mMoveState->addTime(dt);
  }

  // update the animation
  if (mAnimationBlender != 0)  
    mAnimationBlender->addTime(dt);   

  return Static;
}

void HandState::GetNextRemoveSlot(Real& x, Real& z)
{
  // i'm shat at maths - i'm sure there's a *much* better way to do this...  
  int PiecesTaken = mHand->GetNextTakeSlot();
  z = -70 + (20 * int(PiecesTaken / 3)) + (Math::RangeRandom(0, 8) - 4); // the random bit adds a bit of uneven-ness
  while (PiecesTaken > 2) PiecesTaken -= 3;
  x = 110 + (20 * PiecesTaken) + (Math::RangeRandom(0, 8) - 4);  
}
