// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "MovePieceState.h"
#include "Piece.h"
#include "AnimationBlender.h"
#include "Hand.h"

void MovePieceState::Initialise(void) 
{ 
  // record this state type
  mStateType = HandState::MovePiece;   

  // set how long it takes for this part of the move to complete
  mStateDuration = 0.5; 

  // attach the piece to the hand
  AttachPiece("Index3", mMovingPiece, mMPSceneNode);

  // put the hand in it's starting position
  CreateMovementTrack();  

  // set the initial animation
  StartAnimation();  
}

void MovePieceState::CreateMovementTrack()
{    
  if (mActionType == HandState::EnPassant && mEnPassantFirst)
  {
    // get the world coordinates of where the piece is to be moved to    
    GetNextRemoveSlot(mTargetPosition.x, mTargetPosition.z);  
    mTargetPosition.y = -15;
  }
  else
  {
    // get the world coordinates of where the piece is to be moved to
    mTargetPosition = Ogre::Vector3::ZERO;
    Functions::ConvertCoords(mDestinationSquare, mTargetPosition.x, mTargetPosition.z);      
  }    

  // tweak the offsets
  mOffset.x = 32;
  mOffset.y = 50;
  mOffset.z = 62.5;

  // make sure the scene node's rotation is taken into account
  Ogre::Matrix3 nodeRot;      
  mHandNode->_getDerivedOrientation().ToRotationMatrix(nodeRot);
  mOffset = mOffset * nodeRot;

  // tweak the position according to the offsets
  mTargetPosition += mOffset;

  // move the moving piece's node so it's in the right place when we reattach later
  mMPSceneNode->setPosition(mTargetPosition.x - mOffset.x, 0, mTargetPosition.z - mOffset.z);

  // store the movement vector (in world coordinates) 
  // of the target position from the entities current position
  mMovementVector = mTargetPosition - mHandNode->getPosition();  
      
  // create the path to be followed     
  Animation* anim = mSceneManager->createAnimation("HandTrack" + StringConverter::toString(mHand->GetColour()), mStateDuration);

  // Spline it for nice curves
  anim->setInterpolationMode(Animation::IM_SPLINE);

  // Create a track to animate the hand
  NodeAnimationTrack* track = anim->createNodeTrack(0, mHandNode);

  // Setup keyframes
  TransformKeyFrame* key = track->createNodeKeyFrame(0); // startposition
  key->setTranslate(mHandNode->getPosition());
  key->setRotation(mHandNode->getOrientation());  
  
  key = track->createNodeKeyFrame(mStateDuration * 0.5);
  key->setTranslate(Ogre::Vector3(mTargetPosition.x - (mMovementVector.x * 0.5), mTargetPosition.y + 25, mTargetPosition.z - (mMovementVector.z * 0.5)));
  key->setRotation(mHandNode->getOrientation());    

  key = track->createNodeKeyFrame(mStateDuration);
  key->setTranslate(mTargetPosition);  
  key->setRotation(mHandNode->getOrientation()); 
}

void MovePieceState::StartAnimation(void)
{
  // start the movement track
  HandState::BaseStartAnimation();

  // run the move animation
  mAnimationBlender->blend("Move", AnimationBlender::BlendWhileAnimating, mStateDuration * 0.25, false);   
}

HandState::StateType MovePieceState::Update(Real dt)
{         
  // cycle the animation
  HandState::BaseUpdate(dt);  

  // if we reach the end of this state's length, move onto the next one.
  if (mMoveState->getTimePosition() >= mStateDuration)
  {
    // snap the hand into it's final position to make sure
    mHandNode->setPosition(mTargetPosition);

    mSceneManager->destroyAnimation("HandTrack" + StringConverter::toString(mHand->GetColour()));
    mMoveState = 0;

    // go to the next state
    if (mActionType == HandState::Castle && !mCastleFirst)
    {       
      // move to the castle
      mCastleFirst = true;      
      // pick up the second piece
      return HandState::Pickup; 
    }
    else if (mActionType == HandState::EnPassant && !mEnPassantFirst)
    {             
      mEnPassantFirst = true;
      if (mHand->GetColour()) // black
        mSourceSquare = mDestinationSquare - 8;
      else
        mSourceSquare = mDestinationSquare + 8;      
      // pick up the second piece
      return HandState::Pickup; 
    }
    else if (mActionType == HandState::Promotion)
      return HandState::NewPiece; 
    else {
      mCastleFirst = false;
      mEnPassantFirst = false;
      return HandState::SimpleReturn;
    }
  }
  else    
    return mStateType;  
}
