// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "StateManager.h"
#include "State.h"

// Singleton instance
template<> StateManager* Singleton<StateManager>::ms_Singleton = 0;

// Creation/destruction
StateManager::StateManager(RenderWindow* window)
  : mWindow(window), mRunning(true), mRepeatKey(-1),
	mInitialRepeatKeyDelay(0.15), mContinousKeyRepeatDelay(0.05)
{
	OIS::ParamList pl;
	unsigned long hWnd;
	
	// Register frame listener
	Root::getSingleton().addFrameListener(this);

	window->getCustomAttribute("WINDOW", &hWnd);
	pl.insert(ParamList::value_type("WINDOW", StringConverter::toString(hWnd)));
	m_ois = InputManager::createInputSystem(pl);
	
	mMouse = static_cast<Mouse*>(m_ois->createInputObject(OISMouse, true));
	mKeyboard = static_cast<Keyboard*>(m_ois->createInputObject(OISKeyboard, true));
	mMouse->setEventCallback(this);
	mKeyboard->setEventCallback(this);
	const MouseState &ms = mMouse->getMouseState();
	ms.width = window->getWidth();
	ms.height = window->getHeight();
}

StateManager::~StateManager()
{
	// Cleanup all states
	while (!mStateStack.empty())
	{
		mStateStack.back()->cleanup();
		mStateStack.pop_back();
	}

	// Unregister event listeners
	if (mMouse)
		delete mMouse;
	if (mKeyboard)
		delete mKeyboard;
							
	// Destroy event processor
	InputManager::destroyInputSystem(m_ois);

	// Unregister frame listener
	Root::getSingleton().removeFrameListener(this);
}

void StateManager::capture()
{
	mMouse->capture();
	mKeyboard->capture();
}

// State management
void StateManager::changeState(State* state)
{
	// Cleanup current state
	if (!mStateStack.empty())
	{
		mStateStack.back()->cleanup();
		mStateStack.pop_back();
	}

	// Push and initialize new state
	mStateStack.push_back(state);
	mStateStack.back()->initialize();
}

void StateManager::pushState(State* state)
{
	// Pause current state
	if (!mStateStack.empty())
		mStateStack.back()->pause();

	// Push and initialize new state
	mStateStack.push_back(state);
	mStateStack.back()->initialize();
}

void StateManager::popState()
{
	// Cleanup current state
	if (!mStateStack.empty())
	{
		mStateStack.back()->cleanup();
		mStateStack.pop_back();
	}

	// Resume previous state
	if (!mStateStack.empty())
		mStateStack.back()->resume();
}

void StateManager::quit()
{
	mRunning = false;
}

// FrameListener overrides
bool StateManager::frameStarted(const FrameEvent &evt)
{
	// Quit if window is closed
	if (mWindow->isClosed())
		return false;

	// Send key repeated events
	if (mRepeatKey != -1)
	{
		mRepeatKeyDelay -= evt.timeSinceLastFrame;
		while (mRepeatKeyDelay <= 0.0)
		{
			mStateStack.back()->keyRepeated(mRepeatKey);
			mRepeatKeyDelay += mContinousKeyRepeatDelay;
		}
	}

	// Call current state
	mStateStack.back()->frameStarted(evt.timeSinceLastFrame);
	return mRunning;
}

bool StateManager::frameEnded(const FrameEvent &evt)
{
	// Call current state
	mStateStack.back()->frameEnded(evt.timeSinceLastFrame);
	return mRunning;
}

// KeyListener overrides
bool StateManager::keyPressed(const KeyEvent &e)
{
	mStateStack.back()->keyPressed(e);

	// Start repeating key
	mRepeatKey = e.key;
	mRepeatKeyDelay = mInitialRepeatKeyDelay;
	mStateStack.back()->keyRepeated(mRepeatKey);
	return true; 
}

bool StateManager::keyReleased(const KeyEvent &e)
{
	mStateStack.back()->keyReleased(e);

	// Stop repeating key
	if (e.key == mRepeatKey)
		mRepeatKey = -1;
	return true; 
}

void StateManager::setKeyRepeatDelay(Real initial, Real continous)
{
	mInitialRepeatKeyDelay = initial;
	mContinousKeyRepeatDelay = continous;
}

// MouseMotionListener overrides
bool StateManager::mouseMoved(const MouseEvent &e)
{
	mStateStack.back()->mouseMoved(e);
	return true; 
}

bool StateManager::mousePressed(const MouseEvent &e, MouseButtonID id)
{
	mStateStack.back()->mousePressed(e);
	return true; 
}

bool StateManager::mouseReleased(const MouseEvent &e, MouseButtonID id)
{
	mStateStack.back()->mouseReleased(e);
	return true; 
}
